/*
 * Simple program to demonstrate use of the logfile parse engine and
 * indexer provided by parse_log.c and histtee.c.  Paths and hosts are
 * indexed in a case-insensitive manner.
 *
 * Usage:
 *    $ davestat == "$disk:[dir]datestats"
 *    $ davestats filename [flags]
 * 
 * Specify ? as filename to list flags.
 *
 * Author: David Jones
 * Date: 1-JUN-1996
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "parse_log.h"

#include "histtree.h"

int main ( int argc, char **argv )
{
    int length, status, i, j, count, req_status, bytecount, date[6];
    int day_num, good, day_count, ctx, hits, thits, k, enc_val;
    int tally_hours, tally_days, tally_hosts, tally_paths, tally_weekdays;
    int hour_hist[24], weekday_hist[7], daily_hist[31];
    int today, errors, redirects, cache_oks;
    int path_sort;
    int LIB$INIT_TIMER(), LIB$SHOW_TIMER();
    void *host_tree, *path_tree;
    char *fields[8], *flag, *since, *before, *label, *path, *space2;
    char enc[4], cur_day[100], line[1024];
    /*
     * Parse argument list, removing flags from argument list in process.
     */
    tally_hours = tally_days = tally_hosts = tally_paths = 0;
    tally_weekdays = 0;
    path_sort = 2;
    since = before = (char *) 0;
    for ( i = j = 1; i < argc; i++ ) {
	if ( argv[i][0] == '-' ) {
	    flag = &argv[i][1];
	    if ( 0 == strcmp ( flag, "hours" ) ) {
		tally_hours = 1;
	    } else if ( 0 == strcmp ( flag, "hosts" ) ) {
		tally_hosts = 1;
	    } else if ( 0 == strcmp ( flag, "paths" ) ) {
		tally_paths = 1;
	    } else if ( 0 == strcmp ( flag, "weekdays" ) ) {
		tally_weekdays = 1;
	    } else if ( 0 == strcmp ( flag, "directory" ) ) {
		tally_paths = 1;
		path_sort = 0;
	    } else if ( 0 == strcmp ( flag, "since" ) ) {
		i++;
		if ( i < argc ) since = argv[i];
	    } else if ( 0 == strcmp ( flag, "before" ) ) {
		i++;
		if ( i < argc ) before = argv[i];
	    } else {
		fprintf(stderr,"Unknown flag ignored: %s\n", argv[i] );
	    }
	} else if ( argv[i][0] == '?' ) {	/* help */
	    j = 1; break;
 	} else {
	    argv[j++] = argv[i];
	}
    }
    argc = j;
    if ( argc < 2 ) { 
	printf("usage: parse file [-flags]\n"); 
	printf("flags: -hours     Show hour-of-day histogram\n");
	printf("       -paths     Show URL paths, sorted by accesses\n");
	printf("       -weekdays  Show day-of-week histogram\n");
	printf("       -directory Show tree of all accesses\n");
	printf("       -since time  Select only records after 'time'\n");
	printf("       -before time Select only records before 'time'\n");
	exit(1); 
    }
   /*
    * Initialize the statistics counters.
    */
    
    LIB$INIT_TIMER();
    errors = good = redirects = cache_oks = 0;
    today = julian_day ( (int *) 0 );
    if ( argc > 2 ) since = argv[2];
    if ( argc > 3 ) before = argv[3];
    host_tree = create_tree ( '.', 1 );
    path_tree = create_tree ( '/', 0 );
    for ( j = 0; j < 24; j++ ) hour_hist[j] = 0;
    for ( j = 0; j < 7; j++ ) weekday_hist[j] = 0;
    for ( j = 0; j < 31; j++ ) daily_hist[j] = 0;
    /*
     * Tally each file on list.
     */
    for ( i = 1; i < argc; i++ ) {
	status = open_log ( argv[i], since, before );
        if ( !status ) { 
	    printf("error opening '%s'\n", argv[1] ); 
	    continue;
	}
        if ( argc > 2 ) printf("File '%s'...\n", argv[i] );
	cur_day[0] = '\0';
	day_num = day_count = 0;
	/*
	 * Scan records of log file.
	 */
	for ( count = 0; 
	      read_log ( line, fields, &req_status, &bytecount, date ); 
	      count++ ) {
		/* printf("line[%d]: %s", count, line ); */
	    /*
	     * print out mile marker every time day changes.
	     */
	    if ( date[2] != day_num ) {
	        if ( cur_day[0] ) {
		    printf("%s %d\n", cur_day, day_count );
		    day_count = 0;
	        }
	        day_num = date[2];
	        strcpy ( cur_day, fields[3] );
	    }
	    /*
	     * Check the HTTP status returned to the client.
	     */
	    if ( req_status < 300 ) {
	    } else if ( (req_status == 302) || (req_status == 301)) {
		redirects++;
	    } else if ( req_status == 304 ) {
		/* If-modified-since header check passed */
		cache_oks++;
	    } else if ( req_status >= 400 ) {
		errors++;
		continue;
	    }
	    good++;
	    /*
	     * Do other histograms.
	     */
	    if ( tally_hours ) hour_hist[date[3]]++;	/* build hours histogram */
	    if ( tally_weekdays ) {
		j = julian_day ( date );
		weekday_hist[(j+3)%7]++;
	    }
	    day_count++;
	    path = (tally_paths) ? strchr(fields[4],' ') : (char *) 0;

	    if ( path ) {
	        path++;
	        space2 = strchr(path,' ');
	        if ( space2 ) {
		    for ( j = 0; path[j] && path[j]!=' ' &&
		       path[j] != '?' && path[j] != '#'; j++ ) 
		        if ( path[j] == '%' ) {
			    enc[0] = path[j+1]; if ( !enc[0] ) break;
			    enc[1] = path[j+2]; if ( !enc[1] ) break;
			    enc[2] = '\0';
			    sscanf ( enc, "%2x", &enc_val );
			    path[j] = enc_val;
			    for (k=j+1; path[k+2]; k++ ) path[k] = path[k+2];
			    path[k] = '\0';
		        }
		    add_tree ( path_tree, path, j );
	        }
	    }
	    if ( tally_hosts ) {
		add_tree ( host_tree, fields[0], strlen(fields[0]) );
	    }
	}
	close_log();
	printf("%s %d\n", cur_day, day_count );
	printf("total records: %d\n", count );
    }
   LIB$SHOW_TIMER();
    /*
     * display the results.
     */
    printf( "Good records: %d, error records: %d(%5.2f%%)\n", good, errors,
	(100.0*errors)/(errors+good) );
    printf( "Redirects: %d (%5.2f%% of good), cache good: %d ( %5.2f%%)\n",
	redirects, (100.0*redirects)/good, cache_oks,
	(100.0*cache_oks)/good );

    if ( tally_hours ) {
	printf("\nHourly summary:\n");
	for ( j = 0; j < 24; j++ ) if ( hour_hist[j] > 0 ) {
		printf("  %2d: - %d\n", j, hour_hist[j] );
	}
    }

    if ( tally_weekdays ) {
	static char *weekday[] = 
		{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	printf("\nWeekday summary:\n");
	for ( j = 0; j < 7; j++ ) if ( weekday_hist[j] > 0 ) {
		printf("  %s: - %d\n", weekday[j], weekday_hist[j] );
	}
    }

    if ( tally_paths ) {
	printf("\nDirectory tree:\n");
	for  ( ctx = 0; 0 < scan_tree (path_tree, path_sort, &ctx, 
			&label, &hits, &thits ); ) {
	    if ( path_sort ) printf("  /%s:    %d %d\n", label, hits, thits );
	    else {
		if ( hits != thits ) printf("  %*c%s  - %d (%d)\n", 
			ctx*2, '/', label, hits, thits);
		else printf ( "  %*c%s  - %d\n", ctx*2, '/', label, hits );
	    }
	}
    }
    if ( tally_hosts ) {
	printf("\nHosts:\n");
	for ( ctx = 0; 
		0 < scan_tree (host_tree, 0, &ctx, &label, &hits, &thits); ) {
	    if ( thits > 0 ) printf("%s: %d %d\n", label, hits, thits );
	}
    }
    return 1;
}
