/*
 * This module provides routines to building a heirarchal histogram.
 *    void *create_tree ( char *delim, int case_sensitive );
 *		Return pointer to new histogram tree structure.
 *    int add_tree ( void *tree, char *key, int key_len )
 *		Add key to tree structure and/or increment histogram counts.
 *		return value is 1 for success and 0 on error.
 *    int scan_tree ( void *tree, int sort_code, int *context,
 *			char **label, int *hits, int *terminal_hits )
 *		Return over successive calls the data in the tree.
 *		Initialize context to zero for first call.  Sort codes:
 *		0-sort by key, 1-sort by hits.  Label is set to point to
 *		static work area for tree.
 *
 * Author:	David Jones
 * Date:	1-JUN-1996
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "histtree.h"

#define HASH_TABLE_SIZE 4096
#define HASH_MASK 4095
struct treenode {
    struct treenode *sibling;
    struct treenode *children;
    struct treenode *parent;
    struct treenode *next_hash;
    char *name;
    int name_length;
    int hits, terminal_hits;
};
typedef struct treenode *treeptr;
struct treetop {
   struct treenode root;
   char *work;
   treeptr *scan_stack;
   int descendants;			/* number of treenodes malloced */
   int work_size;
   int ss_size;
   int case_sensitive;
    
   char delim;
   treeptr hash_table[HASH_TABLE_SIZE];
};

void *create_tree ( char delim, int case_sensitive )
{
    struct treetop *top;
    int i;

    top = (struct treetop *) malloc ( sizeof(struct treetop) );
    if ( top ) {
        top->root.sibling = top->root.children = (treeptr) 0;
	top->root.parent = top->root.next_hash = (treeptr) 0;
        top->root.hits = top->root.terminal_hits = 0;
        top->root.name_length = 0;
        top->root.name = "";
        top->delim = delim;
        top->case_sensitive = case_sensitive;
        top->descendants = 0;
	top->work_size = 256;
	top->work = malloc ( top->work_size );
	top->ss_size = 0;		/* scan stack size */
	for (i=0; i < HASH_TABLE_SIZE; i++) top->hash_table[i] = (treeptr) 0;
    }

    return (void *) top;
}

/*****************************************************************************/
/*
 * Define hash function for determining initial hash table entry.
 * We assume string has been upcased and that hash table is valid.
 */
#define HASH_THRESHOLD 200
static int hash_name ( char *name, int length )
{
    int hash, i, j;
    char *ptr;

    for ( hash = j = 0, ptr = name; j < length; j++ ) {
	i = (unsigned) *ptr++;
	hash = HASH_MASK & (i*577 ^ hash*3);
    }
    return hash;
}
/*****************************************************************************/

int add_tree ( void *tree, char *name, int name_length )
{
    struct treetop *top;
    treeptr parent, child, sibling;
    char *label, *p;
    int i,j, dlen, count, hash_ndx, test, label_length;
     /*
      * Convert opaque pointer passed from call to local pointer and ensure
      * work buffer at least as big as key string.
      */
    top = (struct treetop *) tree;
    if ( top->work_size <= name_length ) {
	top->work_size = name_length + 256;
	top->work = realloc ( top->work, top->work_size );
    }
    /*
     * Copy key to work area (converting case if caseless) and compute
     * hash table entry position.
     */
    p = top->work;
    if ( name_length > 0 ) if (name[name_length-1] == top->delim) --name_length;
    if ( !top->case_sensitive ) {
	for ( i = 0; i < name_length; i++ ) *p++ = _tolower(name[i]);
        *p = '\0';
    } else {
	strncpy ( p, name, name_length );
	p[name_length] = 0;
    }
    hash_ndx = hash_name ( top->work, name_length );
    /*
     * Scan the hash chain for a match.
     */
    for (child=top->hash_table[hash_ndx]; child; child = child->next_hash)  {
	/*
 	 * work from end of string and check each label for match.
	 */
	sibling = child;
	for ( i = name_length-1; i >= 0; i = i - label_length - 1 ) {
	     label = &top->work[i];
	     for ( j = i; j > 0 && *label != top->delim; --j ) --label;
	     dlen = (*label == top->delim) ? 1 : 0;
	     label_length = i - j + 1 - dlen;
	     if ( label_length != sibling->name_length ) break;
	     test = strncmp ( label+dlen, sibling->name, label_length );
	     if ( test != 0 ) break;

	     sibling = sibling->parent;
	}
	if ( sibling == &top->root && i < 0 ) {
	    /*
	     * We found a match, increment counters.
	     */
	    for ( child->terminal_hits++; child; child = child->parent )
		child->hits++;
	    return 3;
	}
    }
    /*
     * Hash table check was a bust, look to insert position.
     */
    parent = &top->root;
    parent->hits++;

    i = 0;  if ( name[0] == top->delim ) i = 1;

    for ( ; i < name_length; i += label_length + 1) {
	/*
	 * Parse next label.
	 */
	label = &top->work[i];
	for ( j = 0; (j<(name_length-i)) && label[j] != top->delim; j++ );
	label_length = j;
	/*
	 * Scan children of current parent for match.
	 */
	sibling = (treeptr) 0;
	count = 0;
	for ( child = parent->children; 
			child; child = child->sibling ) {
	    count++;
	    test = strncmp ( label, child->name, label_length );
	    if ( test == 0 && (label_length == child->name_length) ) {
		break;
	    } else if ( test < 0 ) {
		/*
		 * Insert new node here.
		 */
		child = (treeptr) 0;
		break;
	    }
	    sibling = child;
	}
	if ( !child ) {
	    /*
	     * allocate new node.
	     */
	    child = (treeptr) malloc ( sizeof(struct treenode) +
		label_length+1 );
	    child->children = child->next_hash = (treeptr) 0;
	    child->name_length = label_length;
	    child->name = (char *) &child[1];
	    strncpy ( child->name, label, label_length );
	    child->name[label_length] = '\0';
	    child->hits = child->terminal_hits = 0;
	    if ( sibling ) {
		child->sibling = sibling->sibling;
		sibling->sibling = child;
	    } else {
		child->sibling = parent->children;
		parent->children = child;
	    }
	    child->parent = parent;
	    top->descendants++;
	}
	/*
	 * Increment counts.
	 */
	child->hits++;
	/*
	 * Child becomes parent for next label.
	 */
	parent = child;
    }
    /*
     * insert into hash table.
     */
    if ( parent->terminal_hits == 0 ) {
	/* first occurence */
	
	if ( parent->next_hash ) printf ( "Multiple hash %d\n", hash_ndx );
	parent->next_hash = top->hash_table[hash_ndx];
	top->hash_table[hash_ndx] = parent;
    }
    parent->terminal_hits++;
    return 1;
}


static int compar_hits ( treeptr *e1, treeptr *e2 )
{
    if ( (*e1)->hits < (*e2)->hits ) return 1;	/* inverted sort */
    else if ( (*e1)->hits == (*e2)->hits ) return 0;
    else return -1;
}
static int compar_thits ( treeptr *e1, treeptr *e2 )
{
    if ( (*e1)->terminal_hits < (*e2)->terminal_hits ) return 1; 
    else if ( (*e1)->terminal_hits == (*e2)->terminal_hits ) return 0;
    else return -1;
}

static int scan_sorted_tree ( void *tree, int sort_key, int *context, 
	char **label, int *hits, int *thits )
{
    struct treetop *top;
    int ctx, i, j, length;
    treeptr *list, cur, parent, child, sibling;
    treeptr path[64];

    top = (struct treetop *) tree;
    ctx = *context;
    /*
     * Initialize if first.
     */
    if ( ctx == 0 ) {
	void qsort();
	printf("sort init, descendants: %d, ss_size: %d\n", top->descendants,
		top->ss_size );
	if ( top->ss_size < top->descendants ) {
	    if ( top->ss_size > 0 ) top->scan_stack = (treeptr *) realloc ( 
		top->scan_stack, sizeof(treeptr) * (top->descendants+1) );
	    else
		top->scan_stack = (treeptr *) malloc ( 
			sizeof(treeptr)*(top->descendants+1) );
	    top->ss_size = top->descendants+1;
	    list = top->scan_stack;
	    list[0] = &top->root;
	    j = 1;
	    for ( i = 0; i < j; i++ ) {
		if ( list[i]->sibling ) list[j++] = list[i]->sibling;
		if ( list[i]->children ) list[j++] = list[i]->children;
	    }

	    qsort ( &list[1], top->descendants, sizeof(treeptr),
		(sort_key == 2) ? compar_hits : compar_thits );
	}
    } else if ( ctx < 0 ) {
	return 0;
    }
    /*
     * advance to next and see if done.
     */
    ctx++;
    if ( ctx > top->descendants ) {
	*context = -1;
	return 0;
    }
    *context = ctx;
    /*
     * Build label.
     */
    cur = top->scan_stack[ctx];
    *hits = cur->hits;
    *thits = cur->terminal_hits;
    for ( i = length = 0, child = cur; child; child = child->parent ) {
	path[i++] = child;
	length = length + 1 + child->name_length;
	if ( child->parent == &top->root ) break;
    }
    j = 0;
    while ( i > 0 ) {
	i--;
	strncpy ( &top->work[j], path[i]->name, path[i]->name_length );
	j += path[i]->name_length;
	if ( i > 0 ) top->work[j++] = top->delim;
    }
    top->work[j] = '\0';
    *label = top->work;
    return ctx;
}

int scan_tree ( void *tree, int sort_key, int *context, 
	char **label, int *hits, int *thits )
{
    struct treetop *top;
    int ctx, i, j;
    treeptr parent, child, sibling;

    if ( sort_key > 1 ) return scan_sorted_tree (
	tree, sort_key, context, label, hits, thits );
    top = (struct treetop *) tree;
    ctx = *context;
    /*
     * advance to next.
     */
    if ( top->ss_size <= ctx ) {
	if ( top->ss_size == 0 ) {
	    top->ss_size = ctx + 16;
	    top->scan_stack = (treeptr *) malloc(sizeof(treeptr) *
		top->ss_size );
	} else {
	    top->ss_size = ctx + 32;
	    top->scan_stack = (treeptr *) realloc(top->scan_stack,
		sizeof(treeptr) * top->ss_size );
	}
    }
    if ( ctx == 0 ) {
	top->scan_stack[0] = &top->root;
	ctx = 1;
    } else if ( ctx < 0 ) return 0;

    parent = top->scan_stack[ctx-1];
    child = parent->children;
    if ( !child ) {
	while ( !child ) {
	    child = parent->sibling;
	    if ( !child ) {
	        ctx--;
		if ( ctx <= 0 ) { *context = -1; return 0; }
		parent = top->scan_stack[ctx-1];
	    }
	    top->scan_stack[ctx-1] = child;
	}
    } else {
	top->scan_stack[ctx++] = child;
    }
    /*
     * Construct label.
     */
    if ( sort_key > 0 ) {
        for ( i = 1, j = 0; i < ctx; i++ ) {
	    strncpy ( &top->work[j], top->scan_stack[i]->name,
		top->scan_stack[i]->name_length );
	    j += top->scan_stack[i]->name_length;
	    if ( i < (ctx-1) ) top->work[j++] = top->delim;
        }
    } else {
	i = ctx - 1;
	j = top->scan_stack[i]->name_length;
	strncpy ( top->work, top->scan_stack[i]->name, j );
    }
    top->work[j] = '\0';
    *label = top->work;
    *hits = child->hits;
    *thits = child->terminal_hits;
    *context = ctx;
    return ctx;
}
